import Clutter from 'gi://Clutter';
import Gio from 'gi://Gio';
import GObject from 'gi://GObject';
import St from 'gi://St';
import * as PanelMenu from 'resource:///org/gnome/shell/ui/panelMenu.js';
import * as PopupMenu from 'resource:///org/gnome/shell/ui/popupMenu.js';
import * as Main from 'resource:///org/gnome/shell/ui/main.js';
import { Extension } from 'resource:///org/gnome/shell/extensions/extension.js';
import GLib from 'gi://GLib';

class ConsoleUtil {
    _available = false;
    _command = '';
    constructor(program, ...args) {
        if (typeof program === 'string') {
            const path = GLib.find_program_in_path(program);
            this._available = Boolean(path);
            if (this._available) {
                this._command = [path, ...args].join(' ');
            }
        }
        if (typeof program === 'object' && Array.isArray(program)) {
            this._available = program.every((file) => GLib.file_test(file, GLib.FileTest.EXISTS));
            if (this._available) {
                this._command = ['cat', ...program, ...args].join(' ');
            }
        }
    }
    async execute(callback) {
        try {
            if (!this.available)
                throw new Error("Util not available");
            const [ok, argv] = GLib.shell_parse_argv(this._command);
            if (!ok || !argv)
                throw new Error("Unable to parse util arguments");
            const p = new Promise((resolve, reject) => {
                let proc = Gio.Subprocess.new(argv, Gio.SubprocessFlags.STDOUT_PIPE | Gio.SubprocessFlags.STDERR_PIPE);
                proc.communicate_utf8_async(null, null, (proc, res) => {
                    try {
                        if (!proc)
                            throw new Error("Util subprocess error");
                        let [, stdout, stderr] = proc?.communicate_utf8_finish(res);
                        if (!proc?.get_successful())
                            throw new Error(stderr);
                        resolve(stdout);
                    }
                    catch (e) {
                        reject(e);
                    }
                });
            });
            callback(await p);
        }
        catch (e) {
            logError(e);
        }
    }
    get available() {
        return this._available;
    }
}

class IbmAcpiUtil extends ConsoleUtil {
    _data = {
        cpu: 0,
        gpu: 0,
        status: '...',
        speed: 0,
        level: '...',
        levels: []
    };
    _hasDedicatedGpu = false;
    constructor() {
        super([
            '/proc/acpi/ibm/thermal',
            '/proc/acpi/ibm/fan'
        ]);
        if (this.available)
            this.update();
    }
    parse(str) {
        const toInt = s => parseInt(s);
        let [temps, status, speed, level, cmd1] = str
            .split(/\n/)
            .map(r => r.slice(r.lastIndexOf('\t') + 1));
        let [cpu, gpu] = temps
            .split(' ')
            .map(toInt);
        speed = parseInt(speed);
        let levels = [];
        const controllable = Boolean(cmd1 && status === 'enabled');
        if (controllable) {
            const [range, ...rest] = cmd1
                .slice(cmd1.indexOf('> is ') + 5, -1)
                .split(', ');
            const [, to] = range
                .split('-')
                .map(toInt);
            const nums = Array.from(Array(to + 1), (_, i) => i);
            const disabled = [0, 'disengaged'];
            levels = [...nums, ...rest]
                .filter(l => !disabled.includes(l))
                .map(l => l.toString());
        }
        if (level === 'disengaged' && Boolean(speed)) {
            level = 'full-speed';
        }
        if (!this._hasDedicatedGpu && gpu > 0) {
            this._hasDedicatedGpu = true;
        }
        if (this._hasDedicatedGpu && gpu <= 0 && this._data.gpu) {
            gpu = this._data.gpu;
        }
        this._data = {
            cpu,
            gpu,
            status,
            speed,
            level,
            levels
        };
    }
    update() {
        return super.execute(this.parse.bind(this));
    }
    setLevel(next) {
        const cmd = `pkexec sh -c "echo level ${next} | tee /proc/acpi/ibm/fan"`;
        const [ok, argv] = GLib.shell_parse_argv(cmd);
        log(`Setting fan level to: ${next}`);
        if (ok && argv?.length) {
            try {
                let proc = Gio.Subprocess.new(argv, Gio.SubprocessFlags.STDOUT_PIPE | Gio.SubprocessFlags.STDERR_PIPE);
                proc.communicate_utf8_async(null, null, (proc, res) => {
                    if (proc) {
                        try {
                            let [, , stderr] = proc?.communicate_utf8_finish(res);
                            if (!proc?.get_successful())
                                throw new Error(stderr);
                        }
                        catch (e) {
                            logError(e);
                        }
                    }
                });
            }
            catch (e) {
                logError(e);
            }
        }
    }
    get cpu() {
        return this._data.cpu.toString();
    }
    get gpu() {
        return this._data.gpu.toString();
    }
    get status() {
        return this._data.status;
    }
    get speed() {
        return this._data.speed.toString();
    }
    get level() {
        return this._data.level;
    }
    get levels() {
        return this._data.levels;
    }
    get dGpu() {
        return this._hasDedicatedGpu;
    }
}

class LsnvmeUtil extends ConsoleUtil {
    _data;
    constructor() {
        super('ls', '-l', '/dev/disk/by-path');
        if (this.available)
            this.update();
    }
    simplify(obj) {
        const nvme = new RegExp(/nvme/i);
        return obj
            .split('\n')
            .filter(l => nvme.test(l))
            .map(r => r
            .slice(r.indexOf('pci-'))
            .replace(/(\.\.\/)/gim, '')
            .split(' -> '))
            .reduce((acc, [path, name]) => {
            const id = 'nvme-pci-' + path.slice(9, 14).replace(':', '');
            acc[name] = id;
            return acc;
        }, {});
    }
    parse(str) {
        this._data = this.simplify(str);
    }
    update() {
        return super.execute(this.parse.bind(this));
    }
    name(key) {
        return this._data[key];
    }
}
class LsblkUtil extends ConsoleUtil {
    _lsnvme;
    _data;
    _hasNvme = false;
    constructor() {
        super('lsblk', '-o', 'HCTL,MODEL,NAME,TRAN,VENDOR,TYPE', '-dnJ');
        if (this.available) {
            this._lsnvme = new LsnvmeUtil();
            this.update();
        }
    }
    simplify(obj) {
        return obj.blockdevices.reduce((acc, { hctl, model, tran, name }) => {
            if (hctl) {
                const [a, b] = hctl.split(':');
                const key = ['drivetemp', 'scsi', a, b].join('-');
                acc[key] = model;
                return acc;
            }
            if (tran === 'nvme') {
                this._hasNvme = true;
                const key = this._lsnvme.name(name);
                acc[key] = model;
                console.log(acc);
                return acc;
            }
            return acc;
        }, {});
    }
    parse(str) {
        this._data = this.simplify(JSON.parse(str));
    }
    update() {
        return super.execute(this.parse.bind(this));
    }
    name(key) {
        return this._data[key];
    }
    hasNvme() {
        return this._hasNvme;
    }
}

class LscpuUtil extends ConsoleUtil {
    _name;
    constructor() {
        super('lscpu');
        if (this.available)
            this.update();
    }
    parse(str) {
        let name = str
            .split(/\n/)
            .find(row => row.includes("Model name:"))
            ?.replace("Model name:", '')
            .split('@')[0]
            ?.replace(/CPU/, ' ')
            .replace(/ +/igm, ' ')
            .trim();
        this._name = name || "Processor";
    }
    update() {
        return super.execute(this.parse.bind(this));
    }
    get name() {
        return this._name;
    }
}

class SensorsUtil extends ConsoleUtil {
    _lscpu;
    _lsblk;
    _data;
    constructor() {
        super('sensors', '-A', '-j');
        if (this.available) {
            this._lscpu = new LscpuUtil();
            this._lsblk = new LsblkUtil();
            this.update();
        }
    }
    simplify(obj) {
        const input = new RegExp(/input$/);
        const fans = new RegExp(/^fan/i);
        return Object.keys(obj).reduce((acc, key) => {
            const inner = Object.keys(obj[key]).find(k => input.test(k));
            if (inner) {
                let value = obj[key][inner];
                if (value > 0 || fans.test(key)) {
                    if (typeof value === 'number')
                        value = value.toString();
                    acc[key] = value;
                }
            }
            else {
                acc[key] = this.simplify(obj[key]);
            }
            return acc;
        }, {});
    }
    organize(data) {
        const cores = new RegExp(/^coretemp/i);
        const drives = new RegExp(/^drivetemp/i);
        const batteries = new RegExp(/^bat/i);
        const tpisa = new RegExp(/^thinkpad-isa/i);
        const temp = new RegExp(/^temp/i);
        const nvme = new RegExp(/^nvme/i);
        return Object.keys(data).reduce((acc, key) => {
            let value = data[key];
            if (Object.keys(value).length === 1) {
                value = Object.values(value)[0];
            }
            if (cores.test(key)) {
                acc.cpu[this._lscpu.name] = value;
                return acc;
            }
            if (drives.test(key)) {
                acc.hdd[this._lsblk.name(key)] = value.toString();
                return acc;
            }
            if (nvme.test(key)) {
                const sensors = Object.keys(value)
                    .filter(v => v.match(/sensor/i));
                value = sensors
                    .map(v => parseFloat(value[v]))
                    .reduce((acc, curr) => acc + curr, 0);
                value /= sensors.length;
                acc.hdd[this._lsblk.name(key)] = value.toString();
                return acc;
            }
            if (batteries.test(key)) {
                const k = key.split('-')[0];
                acc.bat[k] = value.toString();
                return acc;
            }
            if (tpisa.test(key)) {
                const fans = new RegExp(/^fan/i);
                acc.fan = Object.keys(value)
                    .filter(k => fans.test(k))
                    .reduce((acc, k) => {
                    acc[k] = value[k].toString();
                    return acc;
                }, {});
                return acc;
            }
            if (typeof value === 'object') {
                Object.keys(value)
                    .filter(k => temp.test(k))
                    .map(k => {
                    const subkey = [key, k].join('-');
                    acc.other[subkey] = value[k].toString();
                });
                return acc;
            }
            acc.other[key] = value.toString();
            return acc;
        }, { cpu: {}, hdd: {}, bat: {}, fan: {}, other: {} });
    }
    parse(str) {
        this._data = this.organize(this.simplify(JSON.parse(str)));
    }
    update() {
        return super.execute(this.parse.bind(this));
    }
    verify(obj) {
        return Object.keys(obj).length ? obj : false;
    }
    get cpu() {
        return this.verify(this._data.cpu);
    }
    get hdd() {
        return this.verify(this._data.hdd);
    }
    get bat() {
        return this.verify(this._data.bat);
    }
    get fan() {
        return this.verify(this._data.fan);
    }
    get other() {
        return this.verify(this._data.other);
    }
}

const _ = (text) => text;
const UNIT = {
    celsius: "\u00b0C",
    fahrenheit: "\u00b0F",
    rpm: "RPM"
};
const ThermalTitle = GObject.registerClass(class ThermalTitle extends PopupMenu.PopupBaseMenuItem {
    _title;
    constructor(style_class, title) {
        super({
            style_class, reactive: false
        });
        this.setOrnament(PopupMenu.Ornament.HIDDEN);
        this._title = new St.Label({
            text: title,
            x_align: Clutter.ActorAlign.CENTER,
            x_expand: true
        });
        this.add_child(this._title);
    }
});
const ThermalItem = GObject.registerClass(class ThermalItem extends PopupMenu.PopupBaseMenuItem {
    _updater;
    _label;
    _value;
    _unit;
    constructor(style_class, updater, label, props) {
        super({ style_class, reactive: false });
        this._updater = updater;
        if (props?.icon || props?.hideOrnament) {
            this.setOrnament(PopupMenu.Ornament.HIDDEN);
        }
        if (props?.icon) {
            this.add_child(new St.Icon({
                style_class: 'popup-menu-ornament',
                icon_size: 14,
                gicon: props?.icon?.gicon,
                y_align: Clutter.ActorAlign.CENTER
            }));
        }
        this._label = new St.Label({
            style_class: 'label',
            text: label,
            x_align: Clutter.ActorAlign.START,
            x_expand: true
        });
        const value = updater();
        this._value = new St.Label({
            style_class: 'value',
            text: typeof value === 'string' ? value : 'N/A'
        });
        const els = [this._label, this._value];
        if (props?.unit) {
            this._unit = new St.Label({
                style_class: 'unit',
                text: props.unit
            });
            els.push(this._unit);
        }
        els.forEach(el => this.add_child(el));
    }
    update(unit, label) {
        const value = this._updater();
        this._value.set_text(value);
        if (unit)
            this._unit.set_text(unit);
        if (label)
            this._label.set_text(label);
    }
});
const ThermalGroup = GObject.registerClass(class ThermalGroup extends PopupMenu.PopupSubMenuMenuItem {
    _children = [];
    constructor(style_class, updater, label, props) {
        super(label, true);
        this.add_style_class_name(style_class);
        this.setOrnament(PopupMenu.Ornament.HIDDEN);
        if (props?.icon) {
            this.icon.gicon = props?.icon.gicon;
        }
        const data = updater();
        this._children = Object.keys(data).map(key => ({
            key,
            element: new ThermalItem("tpt-popup-submenu-item", () => updater()[key], key, { ...props, icon: false, hideOrnament: true })
        }));
        this._children.forEach(({ element }) => this.menu.addMenuItem(element, undefined));
    }
    update() {
        this._children.forEach(({ element }) => element.update());
    }
});
const ThermalDropDown = GObject.registerClass(class ThermalDropDown extends PopupMenu.PopupSubMenuMenuItem {
    _value;
    _updater;
    _children;
    constructor(style_class, updater, label, items, action) {
        super(label, false);
        this.add_style_class_name(style_class);
        this.setOrnament(PopupMenu.Ornament.HIDDEN);
        this._updater = updater;
        this._value = new St.Label({
            style_class: 'value',
            text: updater(),
            x_align: Clutter.ActorAlign.END,
            x_expand: false
        });
        this.insert_child_at_index(this._value, this.get_children().length - 1);
        this._children = items.map(key => {
            const element = new PopupMenu.PopupBaseMenuItem({
                style_class: 'tpt-popup-dropdown-item'
            });
            element.add_child(new St.Label({
                style_class: 'label',
                text: key,
                x_align: Clutter.ActorAlign.END,
                x_expand: true
            }));
            element.connect("activate", () => action(key));
            return ({ key, element });
        });
        this._children.forEach(({ element }) => this.menu.addMenuItem(element, undefined));
    }
    update() {
        this._value.set_text(this._updater());
    }
});
const IndicatorItem = GObject.registerClass(class IndicatorItem extends St.BoxLayout {
    _updater;
    _icon;
    _value;
    _unit;
    _off;
    constructor(style_class, updater, icon, unit) {
        super({ style_class });
        this._updater = updater;
        this._icon = new St.Icon({
            gicon: icon.gicon,
            icon_size: icon.size,
            style_class: 'icon',
            y_align: Clutter.ActorAlign.CENTER
        });
        this._value = new St.Label({
            text: updater(),
            style_class: 'value',
            y_align: Clutter.ActorAlign.CENTER
        });
        this._unit = new St.Label({
            text: unit,
            style_class: 'unit',
            y_align: Clutter.ActorAlign.CENTER
        });
        this._off = unit === 'RPM' ? new St.Label({
            text: 'OFF',
            style_class: 'unit',
            y_align: Clutter.ActorAlign.CENTER,
        }) : false;
        const els = [this._icon, this._value, this._unit];
        els.forEach(el => this.add_child(el));
        if (this._off)
            this.add_child(this._off);
        this.update();
    }
    off(hide) {
        if (!this._off)
            return;
        this._off.visible = hide;
        this._unit.visible = !hide;
        this._value.visible = !hide;
    }
    update(unit) {
        const v = this._updater();
        this._value.set_text(v);
        if (unit)
            this._unit.set_text(unit);
        this.off(v === "0");
    }
});
const Indicator = GObject.registerClass(class Indicator extends PanelMenu.Button {
    _i = {};
    _tpAcpi;
    _sensors;
    _buttonLayout = new St.BoxLayout({ style_class: 'tpt-button-layout' });
    _bindings = [];
    _elements = ['cpu', 'gpu', 'speed'];
    _updateInterval;
    _temp = UNIT.celsius;
    constructor(ext, align, name) {
        super(align, name);
        this._tpAcpi = new IbmAcpiUtil();
        this._sensors = new SensorsUtil();
        const icons = ['cpu', 'gpu', 'hdd', 'fan', 'sensor'];
        icons.forEach(k => {
            const name = k + '-symbolic.svg';
            const path = [
                ext.dir.get_child('icons').get_path(),
                name
            ].join('/');
            this._i[k] = {
                gicon: Gio.icon_new_for_string(path),
                size: 14
            };
        });
        this._start();
    }
    _checkElements = () => {
        if (parseInt(this._tpAcpi.gpu) <= 0) {
            this._elements = this._elements.filter(e => e !== 'gpu');
        }
    };
    _start = async () => {
        if (this._tpAcpi.available) {
            await this._tpAcpi.update();
            this._checkElements();
            this.appendButton();
            if (this._sensors.available)
                await this._sensors.update();
            this.appendPopupMenu();
        }
        else {
            this._buttonLayout.add_child(new St.Label({ text: "ThinkPad ACPI is not available" }));
        }
        this.add_actor(this._buttonLayout);
        if (this._tpAcpi.available) {
            this._updateInterval = setInterval(this._update.bind(this), 5000);
            this.connect("destroy", this._destroy.bind(this));
        }
    };
    _toFahrenheit = (t) => ((parseFloat(t) * (9 / 5)) + 32).toFixed(1);
    _convert = (t) => {
        if (this._temp === UNIT.fahrenheit) {
            if (typeof t === 'string')
                return this._toFahrenheit(t);
            const o = { ...t };
            Object.keys(o).forEach(k => {
                o[k] = this._toFahrenheit(o[k]);
            });
            return o;
        }
        return t;
    };
    _update = async () => {
        await Promise.all([
            this._tpAcpi.available ? this._tpAcpi.update() : null,
            this._sensors.available ? this._sensors.update() : null
        ].filter(p => p != null));
        this._bindings
            .filter(({ type }) => !['separator', 'tpt-popup-title'].includes(type))
            .forEach(({ element }) => {
            if (typeof element.update === 'function')
                element.update();
        });
    };
    _attach(Element, type, ...args) {
        if (type === 'separator') {
            this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem(''));
            return;
        }
        const element = new Element(type, ...args);
        this._bindings.push({ type, element });
        if (['tpt-button'].includes(type))
            this._buttonLayout.add_child(element);
        if (['tpt-popup-dropdown', 'tpt-popup-submenu', 'tpt-popup-title', 'tpt-popup-item'].includes(type))
            this.menu.addMenuItem(element);
    }
    _destroy() {
        if (this._updateInterval?.get_id()) {
            this._updateInterval.destroy();
        }
    }
    element(key) {
        return this._elements.includes(key);
    }
    appendButton() {
        if (this.element('cpu'))
            this._attach(IndicatorItem, "tpt-button", () => this._convert(this._tpAcpi.cpu), this._i.cpu, this._temp);
        if (this.element('gpu'))
            this._attach(IndicatorItem, "tpt-button", () => this._convert(this._tpAcpi.gpu), this._i.gpu, this._temp);
        if (this.element('speed'))
            this._attach(IndicatorItem, "tpt-button", () => this._tpAcpi.speed, this._i.fan, UNIT.rpm);
    }
    appendPopupMenu() {
        if (this._sensors.available) {
            if (this._sensors.cpu)
                Object.keys(this._sensors.cpu)
                    .forEach(k => this._attach(ThermalGroup, "tpt-popup-submenu", () => this._convert(this._sensors.cpu[k]), k, { icon: this._i.cpu, unit: this._temp }));
            if (this._sensors.hdd)
                this._attach(ThermalGroup, "tpt-popup-submenu", () => this._convert(this._sensors.hdd), "Disks", { icon: this._i.hdd, unit: this._temp });
            if (this._sensors.other)
                this._attach(ThermalGroup, "tpt-popup-submenu", () => this._convert(this._sensors.other), "Thermal", { icon: this._i.sensor, unit: this._temp });
            if (this._sensors.fan)
                this._attach(ThermalGroup, "tpt-popup-submenu", () => this._sensors.fan, "Cooling", { icon: this._i.fan, unit: UNIT.rpm });
        }
        if (this._tpAcpi.available) {
            this._attach(ThermalTitle, "tpt-popup-title", "ACPI");
            if (this.element('cpu')) {
                this._attach(ThermalItem, "tpt-popup-item", () => this._convert(this._tpAcpi.cpu), "CPU", { unit: this._temp, icon: this._i.cpu });
            }
            if (this.element('gpu')) {
                this._attach(ThermalItem, "tpt-popup-item", () => this._convert(this._tpAcpi.gpu), "GPU", { unit: this._temp, icon: this._i.gpu });
            }
            this._attach(ThermalTitle, "tpt-popup-title", "Fan control");
            this._attach(ThermalItem, "tpt-popup-item", () => this._tpAcpi.status, "Status", { hideOrnament: true });
            this._attach(ThermalItem, "tpt-popup-item", () => this._tpAcpi.speed, "Speed", { unit: UNIT.rpm, hideOrnament: true });
            if (this._tpAcpi.status === 'enabled' && this._tpAcpi.levels.length) {
                this._attach(ThermalDropDown, "tpt-popup-dropdown", () => this._tpAcpi.level, "Levels", this._tpAcpi.levels, a => this._tpAcpi.setLevel(a));
            }
            else {
                this._attach(ThermalItem, "tpt-popup-item", () => this._tpAcpi.level, "Level", { hideOrnament: true });
            }
        }
    }
});
class ThinkPadThermal extends Extension {
    _indicator;
    constructor(meta) {
        super(meta);
    }
    enable() {
        this._indicator = new Indicator(this, 0.5, _('ThinkPad Thermal'));
        Main.panel.addToStatusArea(this.uuid, this._indicator);
    }
    disable() {
        this._indicator.destroy();
        this._indicator = null;
    }
}

export { ThinkPadThermal as default };
